extern crate core;

use std::os::raw::{c_char, c_uchar, c_uint};
use std::process::{exit, ExitCode};
use std::ptr::{addr_of, slice_from_raw_parts};
use std::{io, slice, thread};
use std::borrow::BorrowMut;
use std::cmp::max;
use std::env::args;
use std::io::Write;
use std::ffi::c_void;
use std::fmt::format;
use std::fs::{File, OpenOptions};
use std::mem::size_of;
use std::time::{Duration, SystemTime, UNIX_EPOCH};
use clap::Parser;
use x11::keysym::XK_Escape;
use x11::xinput2::XIGrabModeSync;
use x11::xlib::{ButtonPress, ButtonRelease, CurrentTime, GrabModeAsync, GrabModeSync, GrabSuccess, KeyCode, KeyPress, KeyPressMask, KeyRelease, MotionNotify, SyncPointer, Time, XEvent, XFree, XKeyEvent, XKeyPressedEvent, XPointer};
use x11::xrecord::{XRecordAllocRange, XRecordContext, XRecordCreateContext, XRecordDisableContext, XRecordEndOfData, XRecordFreeData, XRecordInterceptData, XRecordStartOfData};
use easymacros::{Instructions, Position};
use easymacros::x11_safe_wrapper::{Keycode, XDisplay};

const DELAY_DEFAULT_MS: u16 = 10;

/// Macro recording module for easymacros. Outputs are partially compatible with xmacro.
#[derive(Parser, Debug)]
#[clap(author, version, about, long_about = None)]
struct Args {
	/// The file to record the macro to. Defaults to writing to stdout.
	#[clap(value_parser, value_name = "output_file", value_hint = clap::ValueHint::FilePath)]
	output_file: Option<std::path::PathBuf>,
	/// Display to run the macro on. This uses the $DISPLAY environment variable by default.
	#[clap(short = 'D', long)]
	display: Option<String>,
	/// Max Delay in milliseconds for macro delays
	#[clap(short, long)]
	max_delay: Option<u64>,
	/// Allow delay capturing in recording output. If this flag is set, the program will ignore the max_delay.
	#[clap(short, long)]
	ignore_delay_capturing: bool,
}

fn main() {
	let args = Args::parse();

	let display = XDisplay::open(args.display.clone());
	let recorded_display = XDisplay::open(args.display.clone());
	let stop_key = get_stop_key(display);
	let writer = MacroWriter::new(args.output_file, args.ignore_delay_capturing);

	event_loop(
		display,
		recorded_display,
		stop_key,
		writer,
		args.max_delay
	);

	display.close();
}

fn get_stop_key(display: XDisplay) -> Keycode {
	let screen = display.get_default_screen();

	let root = display.get_root_window(screen);
	let potential_err = display.grab_keyboard(root, false, GrabModeSync, GrabModeAsync, CurrentTime);

	if potential_err != GrabSuccess {
		eprintln!("Couldn't grab keyboard!");
		exit(1);
	}

	println!("Press the key you want to use to stop recording the macro.");

	let mut stop_key: Keycode = XK_Escape;
	loop {
		display.allow_events(SyncPointer, CurrentTime);
		let ev = XKeyEvent::from(display.window_event(root, KeyPressMask));
		stop_key = ev.keycode;
		break;
	}

	display.ungrab_keyboard(CurrentTime);
	display.ungrab_pointer(CurrentTime);

	stop_key
}

fn event_loop(
	xdpy: XDisplay,
	recdpy: XDisplay,
	stop_key: Keycode,
	mut writer: MacroWriter,
	max_delay: Option<Time>
) {
	// let root = display.get_root_window(screen);
	let protocol_ranges = unsafe { XRecordAllocRange() };

	let pointer_pos: Position<i16> = Position::from(xdpy.query_pointer_pos());

	if pointer_pos != Position (-1, -1) {
		writer.write(Instructions::MotionNotify(pointer_pos))
	}

	let ctx = recdpy.create_record_context(protocol_ranges);
	let ev_cb_data = EvCallbackData::new(writer, xdpy, recdpy, ctx, stop_key, pointer_pos, max_delay);

	if !recdpy.enable_context_async(ctx, Some(ev_callback), addr_of!(ev_cb_data) as *mut c_char) {
		panic!("Failed to enable record context")
	}
	while ev_cb_data.working {
		recdpy.process_replies();
	}

	xdpy.disable_context(ctx);
	xdpy.free_context(ctx);
	unsafe { XFree(protocol_ranges as *mut c_void) };
}

#[repr(C)]
pub struct EvCallbackData {
	pub writer: MacroWriter,
	pub xdpy: XDisplay,
	pub recdpy: XDisplay,
	pub ctx: XRecordContext,
	stop_key: Keycode,
	ev_nr: u32,
	working: bool,
	pos: Position<i16>,
	max_delay: Option<Time>,
	no_keypress_yet: bool,
	last_event: Time,
	moving: bool,
}

impl EvCallbackData {
	pub fn new(
		writer: MacroWriter,
		xdpy: XDisplay,
		recdpy: XDisplay,
		ctx: XRecordContext,
		stop_key: Keycode,
		pos: Position<i16>,
		max_delay: Option<Time>,
	) -> Self {
		EvCallbackData {
			writer,
			xdpy,
			recdpy,
			ctx,
			stop_key,
			ev_nr: 0,
			working: true,
			pos,
			max_delay,
			no_keypress_yet: true,
			last_event: SystemTime::now().duration_since(UNIX_EPOCH).unwrap().as_millis() as Time,
			moving: false,
		}
	}

	pub fn ptr_is_moving(&self) -> bool { self.moving }

	pub unsafe fn update_pos(&mut self, intercept_data: &mut XRecordInterceptData) -> Position<i16> {
		self.pos.0 = *((intercept_data.data as usize + size_of::<i16>() * 10) as *const i16);
		self.pos.1 = *((intercept_data.data as usize + size_of::<i16>() * 11) as *const i16);
		self.pos
	}

	pub fn write_pos(&mut self) {
		self.writer.write(Instructions::MotionNotify(self.pos));
		self.moving = false;
	}

	pub fn maybe_write_delay(&mut self, server_time: Time) {
		if server_time - self.last_event > 1 {
			self.writer.write(Instructions::Delay(calculate_delay(server_time, self.last_event, self.max_delay)));
			self.last_event = server_time;
		}
	}
}

unsafe extern "C" fn ev_callback(closure: *mut c_char, intercept_data: *mut XRecordInterceptData) {
	let data = &mut *(closure as *mut EvCallbackData);
	let intercept_data = &mut *intercept_data;

	if intercept_data.category == XRecordStartOfData {
		println!("Got start of data!");
		data.last_event = intercept_data.server_time;
		XRecordFreeData(intercept_data);
		return;
	} else if intercept_data.category == XRecordEndOfData {
		println!("Got end of data!");
		XRecordFreeData(intercept_data);
		return;
	}

	let ev_type = *(intercept_data.data as *const u8);

	if data.pos.0 == 0 || data.pos.1 == -1 {
		if ev_type == MOTIONNOTIFY_U8 {
			data.update_pos(intercept_data);
			data.write_pos();
		} else {
			println!("Move your cursor so the macro can start with a fixed cursor position!
			Skipping event...");
		}
	} else if data.no_keypress_yet && ev_type == KEYRELEASE_U8 {
		println!("Skipping KeyRelease without recorded KeyPress...");
	} else {
		// if
		match ev_type {
			MOTIONNOTIFY_U8 => {
				data.update_pos(intercept_data);

				if !data.moving {
					data.moving = true;
				}
			}
			KEYPRESS_U8 | KEYRELEASE_U8 => {
				let kc: u8 = *((intercept_data.data as usize + 1) as *const u8);
				let keyname = data.xdpy.keycode_to_string(kc as u32);

				if ev_type == KEYPRESS_U8 && kc == data.stop_key as u8 {
					data.working = false;
				} else {
					if ev_type == KEYPRESS_U8 {
						data.no_keypress_yet = false;
					}

					data.maybe_write_delay(intercept_data.server_time);

					if data.ptr_is_moving() { data.write_pos(); }

					data.writer.write(if ev_type == KEYPRESS_U8 {
							Instructions::KeyStrPress(keyname)
						} else {
							Instructions::KeyStrRelease(keyname)
						});
				}
			}
			BUTTONPRESS_U8 | BUTTONRELEASE_U8 => {
				let bc: u8 = *((intercept_data.data as usize + 1) as *const u8);

				data.maybe_write_delay(intercept_data.server_time);

				if data.ptr_is_moving() { data.write_pos(); }

				if ev_type == BUTTONPRESS_U8 {
					data.writer.write(Instructions::ButtonPress(bc));
				} else {
					data.writer.write(Instructions::ButtonRelease(bc));
				}
			}
			_ => eprintln!("Unknown event type: {:?}", ev_type)
		}
	}

	data.ev_nr += 2;
	XRecordFreeData(intercept_data)
}

fn calculate_delay(server_time: Time, last_event: Time, max_delay: Option<Time>) -> Time {
	if let Some(max) = max_delay {
		let max = max as u64;
		let delay = server_time - last_event;

		if delay > max {
			max
		} else {
			delay
		}
	} else {
		server_time - last_event
	}
}

pub struct MacroWriter {
	outfile: Box<dyn Write>,
	ignore_delay_capturing: bool,
}

impl MacroWriter {
	pub fn new(outfile: Option<std::path::PathBuf>, ignore_delay_capturing: bool) -> Self {
		Self {
			outfile: if let Some(outfile) = outfile {
				Box::new(File::create(outfile).expect("Failed to create output file"))
			} else {
				Box::new(io::stdout())
			},
			ignore_delay_capturing,
		}
	}

	pub fn write(&mut self, instruction: Instructions) {
		if self.ignore_delay_capturing {
			if let Instructions::Delay(_) = instruction {
				return;
			}
		}

		writeln!(&mut self.outfile, "{}", instruction).expect("Failed to write instruction to outfile");
	}
}

const KEYPRESS_U8: u8 = KeyPress as u8;
const KEYRELEASE_U8: u8 = KeyRelease as u8;
const BUTTONPRESS_U8: u8 = ButtonPress as u8;
const BUTTONRELEASE_U8: u8 = ButtonRelease as u8;
const MOTIONNOTIFY_U8: u8 = MotionNotify as u8;

